/*************************************************************************
*                                                                        *
*          AutoRun: Automatically display CD-ROM documentation           *
*                   Copyright (C) 2002, John Zaitseff                    *
*                                                                        *
*************************************************************************/

/*  Author:    John Zaitseff <J.Zaitseff@zap.org.au>
    Date:      16th June, 2002
    Version:   2.2

    This program uses the file AutoRun.inf to automatically display the
    contents of any documentation.  It has been written to provide a simple
    (but suitable) executable for the AutoRun feature of Microsoft Windows.
    Please see the file README.txt for more information.


    This program is free software, and may be distributed and/or modified
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either Version 2 of the License or any later
    version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (in the file License.txt); if not, write to
    the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
    Boston, MA 02111-1307, USA.

    The file License.txt also provides a special exception to the GNU
    General Public License: you may choose to distribute only the files
    AutoRun.exe, AutoRun.inf and AutoRun.ico, and place a pointer (in the
    form of a publicly-accessible URL) to the source code in the file
    AutoRun.inf.  Please see License.txt and README.txt for more information.
*/


#include <stdarg.h>

#define STRICT
#define WIN32_MEAN_AND_LEAN
#include <windows.h>

#include "AutoRun.h"


/*  ----------------------------------------------------------------------
 *  Global constants and variables
 */

#define MAX_STRING 1024
    /* Maximum number of characters (not bytes!) in a string buffer */

#define NUL 0
    /* ASCII NUL character, used as an end of string marker */

LPCTSTR window_class_name = "AutoRun";
    /* Window class name, for the invisible window */

#ifdef __DEBUG
LPCTSTR exe_file_extension = ".exe";
#endif
#define exe_file_extension_len 4
    /* File extension of the AUTORUN.EXE file */

LPCTSTR inf_file_extension = ".inf";
#define inf_file_extension_len 4
    /* File extension of the AUTORUN.INF file */

LPCTSTR cur_dir = ".\\";
#define cur_dir_len 2;
    /* Current directory */

LPCTSTR inf_section = "display";
    /* Section in the AUTORUN.INF file that contains files to display */

LPCTSTR inf_key_prefix = "display";
#define inf_key_prefix_len 7
#define inf_key_len        9  /* Allow for two extra characters */
    /* Partial key in the [display] section for each file to display */

#define display_count_first '1'
#define display_count_last  '9'
    /* First and last values for the "displayN = ..." key-value pair.  These
       MUST be single ASCII characters */


HINSTANCE hInst;
    /* The instance of this application */

HWND hWnd;
    /* Handle to the application's window */

TCHAR caption[MAX_STRING];
    /* Caption for dialog boxes */


/*  ----------------------------------------------------------------------
 *  Macro definitions
 */

#ifdef __DEBUG
#  define ASSERT(c) \
        if (! (c)) {                                                       \
            MessageBox( NULL, "Assertion failed:\n" #c, window_class_name, \
                        MB_OK | MB_ICONSTOP);                              \
            ExitProcess(2);                                                \
        }
#else /* ! __DEBUG */
#  define ASSERT(c)
#endif /* ! __DEBUG */


/*  ----------------------------------------------------------------------
 *  Function prototypes
 */

int APIENTRY WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpCmdLine, int nShowCmd);
int RetrieveDisplayFileName (LPCTSTR inf_filename, LPCTSTR inf_section,
                             LPCTSTR inf_key, LPCTSTR app_directory,
                             LPTSTR display_filename);
void ShowOurError (UINT err_ids, ...);
void ShowSysError (LPCTSTR pathname, DWORD err);


/*  ----------------------------------------------------------------------
 *  Function:   WinMain        - Main entry point for Windows application
 *  Parameters: hInstance      - Handle to the current instance
 *              hPrevInstance  - Handle to previous instance (always 0)
 *              lpCmdLine      - Pointer to command line parameter
 *              nShowCmd       - How the window is to be shown
 *  Returns:    int            - Return code to the operating system
 *
 *  This function is the main entry point for all Windows applications.
 *  It usually contains a message loop to deal with the graphical interface.
 *  In this application, however, almost everything that needs to be done is
 *  done here, to display an appropriate HTML file.
 *
 *  Note that lpCmdLine is always NULL as it is not initialised by the
 *  minimalist c0w32.asm start-up file used by this application.
 */

#pragma argsused
int APIENTRY WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpCmdLine, int nShowCmd)
{
    TCHAR app_filename[MAX_PATH];       /* The full pathname of application */
    TCHAR app_directory[MAX_PATH];      /* Directory component */
    TCHAR *app_filepart;                /* Pointer to within app_filename */
    TCHAR inf_filename[MAX_PATH];       /* Full pathname of INF file */
    TCHAR inf_key[inf_key_len];         /* Key to retrieve from INF file */
    TCHAR *p_inf_key;                   /* Pointer to within the INF key */
    TCHAR display_filename[MAX_PATH];   /* Full pathname of file to display */
    TCHAR display_count;                /* Count of display files seen */
    TCHAR not_assoc;                    /* which file was not associated */
    TCHAR buf[MAX_PATH];                /* Temporary buffer */

    WNDCLASS wc;                        /* Window class for application */
    HANDLE fh;                          /* FindFirstFile handle */
    WIN32_FIND_DATA fdat;               /* FindFirstFile data block */
    DWORD err;                          /* Value from GetLastError() */

    int n;


    /* Preliminary initialisation */
    hInst = hInstance;

    /* Load the application's caption, for use in MessageBox() */
    LoadString(hInstance, IDS_CAPTION, caption, MAX_STRING);
    ASSERT(*caption != NUL);

    /* Register the application's window class so that the icon will appear
       correctly in MessageBox() */
    wc.style = 0;
    wc.lpfnWndProc = DefWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, "MAINICON");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) COLOR_WINDOW;
    wc.lpszMenuName = NULL;
    wc.lpszClassName = window_class_name;

    if (! RegisterClass(&wc)) {
        ShowSysError(NULL, GetLastError());
    }

    /* Create a hidden window (1x1 pixel) with the correct icon */
    if ((hWnd = CreateWindow(window_class_name, window_class_name, 0, 0, 0,
                             1, 1, NULL, NULL, hInstance, NULL)) == NULL) {
        ShowSysError(NULL, GetLastError());
    }
    ShowWindow(hWnd, SW_HIDE);

    /* Get the application's full filename */
    GetModuleFileName(NULL, buf, MAX_PATH);
    GetFullPathName(buf, MAX_PATH, app_filename, &app_filepart);
    ASSERT(app_filepart != app_filename);

    /* Get the directory of the application, from its full filename */
    n = app_filepart - app_filename;
    ASSERT(app_filepart - app_filename > 0);
    lstrcpy(app_directory, app_filename);
    app_directory[n] = NUL;

    /* Determine the name of the INF file: replace the EXE file extension
       with the INF file extension */
    lstrcpy(inf_filename, app_filename);
    n = lstrlen(inf_filename) - exe_file_extension_len;
    ASSERT(lstrlen(inf_filename) - exe_file_extension_len > 0);
    ASSERT(lstrcmpi(inf_filename + n, exe_file_extension) == 0);
    lstrcpy(inf_filename + n, inf_file_extension);

    /* Check if the INF file actually exists */
    if ((fh = FindFirstFile(inf_filename, &fdat)) == INVALID_HANDLE_VALUE) {
        if ((err = GetLastError()) == ERROR_FILE_NOT_FOUND) {
            /* The INF file was not found */
            ShowOurError(IDS_MISSING_INF, inf_filename);
        } else {
            /* Some other problem with the INF file occurred */
            ShowSysError(inf_filename, err);
        }
    } else {
        /* The INF file was found, as expected */
        FindClose(fh);
    }

    /* Initialise the main loop */
    not_assoc = 0;

    ASSERT(lstrlen(inf_key_prefix) + 2 <= sizeof(inf_key) / sizeof(TCHAR));
    lstrcpy(inf_key, inf_key_prefix);
    p_inf_key = inf_key + inf_key_prefix_len;
    *(p_inf_key + 1) = NUL;     /* Make sure there will be an end of string */

    /* Now, at long last, try to find the appropriate file and display it */

    for (display_count = display_count_first;
         display_count <= display_count_last;
         display_count++) {

        /* Create the key string (to find "displayN = file" in the INF file */
        *p_inf_key = display_count;           /* display_count is in ASCII */

        /* Retrieve the fully-qualified filename from the INF file */
        if (! RetrieveDisplayFileName(inf_filename, inf_section, inf_key,
                                      app_directory, display_filename)) {
            /* The key was not found in the INF file: end the loop */
            break;
        }

        /* Now, see if the file actually exists */
        if ((fh = FindFirstFile(display_filename, &fdat))
                == INVALID_HANDLE_VALUE) {
            if ((err = GetLastError()) == ERROR_FILE_NOT_FOUND) {
                /* The file was not found: try another file */
                ;
            } else {
                ShowSysError(display_filename, err);
            }
        } else {
            /* The file was found! */
            FindClose(fh);

            /* Try to execute the file using the file associations registered
               with Windows */
            if ((int) ShellExecute(NULL, "open", display_filename, NULL, NULL,
                                   SW_SHOWDEFAULT) > 32) {
                /* All OK: the index file has been opened.  Note that the API
                   documentation states that SW_SHOWDEFAULT should be replaced
                   with a zero, but this is incorrect */
                ExitProcess(0);
            } else {
                if ((err = GetLastError()) == ERROR_NO_ASSOCIATION) {
                    /* The file does not have a corresponding application that
                       can display/execute it: try another file, but also
                       remember which file was not associated */
                    not_assoc = display_count;
                } else {
                    ShowSysError(display_filename, err);
                }
            }
        }

        /* At this point in the loop, either the file was not found, or the
           file is not associated with any application.  In either case, try
           another file */
    }

    /* None of the files listed in the INF file could be displayed */

    /* Check if at least "display1" was specified */
    if (display_count == display_count_first) {
        ShowOurError(IDS_MISSING_KEY, inf_filename, inf_section, inf_key);
    }

    /* Maybe one of the files was not associated with any application (the
       rest of the files were not found) */
    if (not_assoc != 0) {
        if (not_assoc != (display_count - 1)) {
            /* This is very subtle: only reload display_filename for the key
               in not_assoc if it is not already there */
            *p_inf_key = not_assoc;
            RetrieveDisplayFileName(inf_filename, inf_section, inf_key,
                                    app_directory, display_filename);
        }
        ShowOurError(IDS_NOT_ASSOC, display_filename);
    }

    /* None of the files were found at all! */
    ShowOurError(IDS_NONE_FOUND, inf_filename);

    return 1;       /* Dummy return, not reached */
}


/*  ----------------------------------------------------------------------
 *  Function:   RetrieveDisplayFileName  - Get full filename from INF file
 *  Parameters: inf_filename             - INF full filename
 *              inf_section              - Section in INF file "[display]"
 *              inf_key                  - Key in INF file "displayN"
 *              app_directory            - Directory of the application
 *              display_filename         - (returned) full filename
 *  Returns:    int                      - TRUE if display_filename retrieved
 *
 *  This function reads a value from the INF file (section in inf_section,
 *  key in inf_key) and makes it into a full filename.  TRUE is returned if
 *  display_filename was correctly set, FALSE if the key could not be found
 *  in the INF file.
 */

int RetrieveDisplayFileName (LPCTSTR inf_filename, LPCTSTR inf_section,
                             LPCTSTR inf_key, LPCTSTR app_directory,
                             LPTSTR display_filename)
{
    TCHAR buf[MAX_PATH];


    /* Retrieve the value of the key (in inf_key) from the INF file */
    if (GetPrivateProfileString(inf_section, inf_key, "", buf,
                                MAX_PATH, inf_filename) <= 0) {
        /* The key was not found in the INF file */
        return FALSE;
    }

    /* Make the retrieved filename fully-qualified */
    if (buf[0] == '\\' || buf[1] == ':') {
        /* Specified filename starts with "\" or "X:" (this should never
           occur, but we do allow it...) */
        lstrcpy(display_filename, buf);
    } else {
        /* Append the specified filename to the application's directory */
        ASSERT(lstrlen(app_directory) + lstrlen(buf) <= MAX_PATH);
        lstrcpy(display_filename, app_directory);
        lstrcat(display_filename, buf);
    }

    return TRUE;
}


/*  ----------------------------------------------------------------------
 *  Function:   ShowOurError  - Show a custom error message
 *  Parameters: err_ids       - Identifier for LoadString()
 *              ...           - Additional parameters
 *  Returns:    (nothing)     - This function does NOT return!
 *
 *  This function loads the string identified by err_ids, parses it through
 *  FormatMessage and displays it in a dialog box.  Note that this function
 *  terminates the program via a call to ExitProcess()!
 */

void ShowOurError (UINT err_ids, ...)
{
    va_list ap;                 /* Pointer to list of arguments */
    TCHAR msg[MAX_STRING];      /* Complete error message */
    TCHAR err[MAX_STRING];      /* Loaded error string, from err_ids */


    va_start(ap, err_ids);

    LoadString(hInst, err_ids, err, MAX_STRING);
    ASSERT(lstrlen(err) != 0);

    FormatMessage(FORMAT_MESSAGE_FROM_STRING, err, 0, 0, msg, MAX_STRING, &ap);

    va_end(ap);

    MessageBox(hWnd, msg, caption, MB_OK | MB_ICONSTOP);
    ExitProcess(1);
}


/*  ----------------------------------------------------------------------
 *  Function:   ShowSysError  - Show a system error message
 *  Parameters: pathname      - Filename causing the error
 *              err           - The actual error code from GetLastError()
 *  Returns:    (nothing)     - This function does NOT return!
 *
 *  This function displays the message corresponding to the system error in
 *  err in a dialog box, along with the name of the file causing the problem.
 */

void ShowSysError (LPCTSTR pathname, DWORD err)
{
    TCHAR msg[MAX_STRING];      /* Complete error message */
    TCHAR sys[MAX_STRING];      /* Error message from the system */


    if (! pathname) {
        *msg = NUL;
    } else {
        lstrcpy(msg, pathname);
        lstrcat(msg, ": ");
    }

    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, err, 0, sys,
                  MAX_STRING, NULL);
    ASSERT(lstrlen(msg) + lstrlen(sys) < MAX_STRING);
    lstrcat(msg, sys);

    MessageBox(hWnd, msg, caption, MB_OK | MB_ICONSTOP);
    ExitProcess(1);
}
